import ij.*;
import ij.plugin.*;
import ij.measure.ResultsTable;
import ij.plugin.frame.Recorder;
import ij.process.*;
import ij.gui.*;
import ij.io.*;
import ij.text.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

/**
 * This ImageJ plugin (the "Plugin") closes all image windows and non-image windows 
 * as selected by the user.  It bases on the   Dispose_All_Windows   plugin
 * by Tony Collins and J. Anthony Parker (using their version of March 8, 2004),
 * Recently added features and improvements include: 
 * 	- splitting the dialog wondow into several pages if needed.  
 *			This applies if there are more items to be listed 
 *			than space is avalible for to list them in a single page
 * 	- optionally closing non-image windows.   
 *			This option may be useful if may text windows are open. 
 * 			The active "ResultsTable" and the "Recoder" window however 
 * 			won't  be disposed because closing the window won't kill the
 *			underlying objects but make them inaccessible.
 *
 * To quickly close all image or non-image windows without selecting or deselcting 
 * individual ones, press the <SHIFT> key at starting the plugin.
 * To close all image windows without any dialog or confirmation, press the <ALT> key.
 * Note that uisng this button here may be inpracticable in Microsoft Windows(tm) computers.
 * 
 * If the Plugin is called from the Macro executer with options set, it will dispose
 * without confirmation any image and non-image windows as specified
 * (regardles it is locked or changes are not saved.)
 * Examples:
 *	run("Dispose All Windows", "image_1.tif image_2.tif image_3.tif Exception");
 *		disposing the specified windows, (which could be either the title of
 *		images like in this example "image_1.tif",  and/or titles of non-image windows
 *		like, in this example, the title of the text window "Exeption".)  
 *		Note that spaces in window names have to be replaced by underscores.
 *		If there are two windows with the same (space to underscore replaced) title,
 *		add an underscore for referring the second one.
 *		Window names that doesn't match with an existing window
 *		will be ignored.
 *	run("Dispose All Windows", "/all images");
 *		dispose all image windows without specifying some among them.
 *	run("Dispose All Windows", "/all non-images");
 *		dispose all non-image windows.
 *	run("Dispose All Windows", "/all images non-images");
 *		dispose all image and non-image windows.
 * These options and commands may be combind freely.
 * For displaying the user dialog, call the plugin from the Macro executer without options set.
 * Example:
 *	run("Dispose All Windows");
 * 
 * 
 * Find updates at 
 *	http://www.neuroscience.ubc.ca/faculty/murphy_software.html#dispose_all
 * The plugin is publicly licenced according to the most recent GNU General Public License
 * 	(e.g. http://www.gnu.org/licenses/gpl-3.0.txt ).
 * Usage on own risk.
 * Please don't hesitate sending any comments, suggestions, bug reports, swears, etc. to Albrecht Sigler.
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @author Tony Collins, PhD <tonyc@uhnresearch.ca>
 * @author Albrecht Sigler, PhD <asigler@interchange.ubc.ca>, <asigler@gwdg.de>
 * @version updated  January 5 2013 by Albrecht Sigler
 *  
 */


public class Dispose_All_Windows implements PlugIn {
	protected ImageJ ij;
	private GenericDialog gd = null;
	private Button selectAllButton = null, deselectAllButton = null;
	private ActionListener gdal = new ActionListener(){
		public void actionPerformed( ActionEvent e ) {
			int index = -1, indices = 2;
			boolean select = selectAllButton == e.getSource();
			if (select) {
				java.util.Vector ch = gd.getChoices();
				Choice methodChoice = (Choice) ch.get(0);
				if (null != methodChoice) {
					index = methodChoice.getSelectedIndex();
					indices = methodChoice.getItemCount();
				}
			}
			java.util.Vector cb = gd.getCheckboxes();
			for (int i = 0; i < cb.size(); i++) {
				Checkbox cbx = (Checkbox) cb.get(i);
				if (null == cbx || gd.wasCanceled()) return;  // -- crash prevention
				if (cbx.isEnabled()) {
					String label = cbx.getLabel();
					// -- don't select "Memory" window because the user presumeably wants to keep it
					boolean isMemoryWin = 0 == label.compareTo("Memory");
					if (0 == label.compareTo("_")) {  // --  exclude putatively invalid list entries
						cbx.setState(false);
						cbx.setEnabled(false);
					} else if (select && 3 < indices && 1 == index) {
						if (label.endsWith("*")) cbx.setState(false);
						else cbx.setState(!isMemoryWin);
					} else cbx.setState(select && !isMemoryWin);
				}
			}
		}
	};
	private static boolean cbHeightMeasured = false;
	private static int checkboxHeight = 0;
	
	//  ---------  OPTIONS  ------------------
	private final boolean dontQuitAtCancel = true;   // -- look & feel of the generic dialog "Cancel" button
	//		boolean "dontQuitAtCancel"  applies only to cases if the "Skip" button is not avaliable
	//		(i.e. in older versions of ImageJ).  If "dontQuitAtCancel"  is set true, 
	//		pressing the "Cancel" button will be interpreted like pressing the "Skip" button.
	//		Otherwise pressing the "Cancel" button would cause the plugin to be quitted before disposing anything.
	private final boolean reallySilentlyAtAltKey = true;   // -- causes the plugin to close all windows without any dialog
	//		for compatibliy to preceeding versions of this plugin.
	//		Otherwise the <ALT> key would be interpreted like the <SHIFT> key  (display short dialog).
	private final boolean ignoreNonImageWindows = false;   // -- set it true if your don't like this nice feature.
	//  ---------   ---------   --------------------
	
	public void run(String args) {
		boolean macroRunningSilently = IJ.macroRunning();
		if (macroRunningSilently) {
			String mArgs = Macro.getOptions();
			if (null == mArgs) macroRunningSilently = false;
			else args = mArgs;
		}
		if (null == args) args = "";
		else args = args.trim();
		
		//  --  get collection of IMAGE windows --
 		int[] wList = WindowManager.getIDList();
		if (wList == null) wList = new int[0];
		ImagePlus [] imp = new ImagePlus [wList.length];
		for (int i=0; i < wList.length; i++) {
			imp[i] = WindowManager.getImage(wList[i]);
			if (null != imp[i]) imp[i].setTitle(imp[i].getTitle()); // -- as in WindowManager.repaintImageWindows();
		}
		if (0 < wList.length && 12 > wList.length) IJ.wait(20);
		boolean [] imgChanges = new boolean [wList.length];   // -- true if image possesses unsaved changes
		boolean anyChanges = false;    // -- true if there is at least one image with unsaved changes
		for (int i=0; i < wList.length; i++)  {
			if (null != imp[i]) imgChanges[i] = imp[i].changes;
			if (imgChanges[i]) anyChanges = true;
		}
		String [] imgLabels = new String[wList.length];  // -- labels of checkboxes in dialog
		for (int i=0; i < imgLabels.length; i++) {
			if (null != imp[i]) imgLabels[i] = imp[i].getTitle();
			if (null != imgLabels[i]) imgLabels[i] = imgLabels[i].replaceAll(" ","_");  // replace all spaces in labels by underscores
			else imgLabels[i] = "_";
			if (0 == imgLabels[i].length()) imgLabels[i] = "_";
			if (54 < imgLabels[i].length()) imgLabels[i] = imgLabels[i].substring(0, 50)+"...";
			boolean isUnique = true;  // -- true if there is no other Image with same label
			do {
				isUnique = true;
				// -- find image windows above this one with same label
				for (int ii = 0; ii < i; ii++) 
					if (isUnique)
						if (0 == imgLabels[i].compareToIgnoreCase(imgLabels[ii]))
							isUnique = false;
				// -- rename label, if needed, in order to making it unique
				if (!isUnique) {
					if (imgLabels[i].endsWith("...")) imgLabels[i]+=".";
					else imgLabels[i] +="_";
				}
			} while (!isUnique);
		}
		boolean shiftKeyIsDown = false;
		try {shiftKeyIsDown = IJ.shiftKeyDown();}
		catch (NoSuchMethodError e) {}
		boolean altKeyIsDown = false;
		try {altKeyIsDown = IJ.altKeyDown();}
		catch (NoSuchMethodError e) {}
		
		//  --  get collection of NON-IMAGE windows --
		Frame [] frames={};
		if ((!altKeyIsDown || !reallySilentlyAtAltKey) && !ignoreNonImageWindows) {
			try {
				frames=WindowManager.getNonImageWindows();
			}
	 		catch (NoSuchMethodError ex) {}  // -- older versions of ImageJ don't have this method
 			catch (IllegalAccessError ex) {}
		}
		boolean [] shouldBeKept = new boolean [frames.length];
		String [] frameLabels = new String[frames.length];
		for (int i=0; i < frameLabels.length; i++) {
			frameLabels[i] = frames[i].getTitle();
			if (null != frameLabels[i]) frameLabels[i] = frameLabels[i].replaceAll(" ","_");
			else frameLabels[i] = "";
			if (0 == frameLabels[i].length()) {
				frameLabels[i] = "_";
				shouldBeKept[i] = true;
			}
			if (54 < frameLabels[i].length()) frameLabels[i] = frameLabels[i].substring(0, 50)+"...";
			boolean isUnique = true; // -- true if there isn't another window with same label
			do {
				isUnique = true;
				// -- find image windows with same label
				for (int ii = 0; ii < imgLabels.length; ii++) 
					if (0 == frameLabels[i].compareToIgnoreCase(imgLabels[ii])) {
						isUnique = false;
						break;
					}
				// -- find non-image windows (prior to this one) with same label 
				if (isUnique) for (int ii = 0; ii < i; ii++) 
					if (0 == frameLabels[i].compareToIgnoreCase(frameLabels[ii])) {
						isUnique = false;
						break;
					}
				// -- rename label if it's not unique
				if (!isUnique) {
					if (frameLabels[i].endsWith("...")) frameLabels[i]+=".";
					else frameLabels[i] +="_";
				}
			} while (!isUnique);
			// -- identify and exclude the active results window (which should't be disposed)
			if (IJ.isResultsWindow() ) {
				if (frames[i] instanceof TextWindow)  {
					TextPanel textPanel = ((TextWindow)frames[i]).getTextPanel();
					if (textPanel == IJ.getTextPanel()) shouldBeKept[i] = true;
				}
			}
			// -- identify and exclude macro recorder window (which should't be disposed also)
			if (!shouldBeKept[i])
				if (frames[i] instanceof Recorder)
					shouldBeKept[i] = true;
		}
		// -- check if there is at least one window that could be closeed
		if (0 < frames.length) {
			boolean thereIsOne = false;
			for (int i=0; i < frames.length; i++) {
				thereIsOne = !shouldBeKept[i];
				if (thereIsOne) break;
			}
			if (!thereIsOne) {
				frames = new Frame[0];
				shouldBeKept = new boolean [0];
				frameLabels = new String[0];
			}
		}
		
		boolean allimages = false;    // -- true if user selected to dispose all images
		boolean allframes = false;    // -- true if selected to dispose non-image windows
		boolean [] imgToClose = new boolean [wList.length];   // -- true if user selected an image for closing
		boolean [] winToClose = new boolean [frames.length];   // -- true if user selected a non-image window for disposing
		boolean optionalCloseAvailabe = false;    // -- optionally the user can skip any confirmation requests
				// 	before disposing a text window with unsaved changes.
				// 	This option is available in ImageJ version 1.43s and newer.
		boolean [] confirmClosing = new boolean [frames.length];    // -- true if user wants to confirm
				// 	disposing the item if it is a text window with unsaved changes
				// 	(with ImageJ version 1.43s and newer, for whose this option is available.)
		
		//  -----------------------------------------------------------------------------
		
		if (macroRunningSilently) {  // true if macro is running WITH options set
			
			if (0 == args.length()) return;
			String [] cmds = {};
			cmds = args.split("\\ ");  // -- array of macro commands to interpret
			int c = -1;
			while (++c < cmds.length)  if (0 < cmds[c].length()) {
				if (0 == cmds[c].compareToIgnoreCase("/all")) {
					int cc = c;
					while (++cc < cmds.length)  if (0 < cmds[cc].length()) {
						if (0 == cmds[cc].compareToIgnoreCase("image")) {
							allimages = true;
							c++;
							if (allframes) cc = cmds.length;
						} else if (0 == cmds[cc].compareToIgnoreCase("images")) {
							allimages = true;
							c++;
							if (allframes) cc = cmds.length;
						} else if (0 == cmds[cc].compareToIgnoreCase("non-image")) {
							allframes = true;
							c++;
							if (allimages) cc = cmds.length;
						} else if (0 == cmds[cc].compareToIgnoreCase("non-images")) {
							allframes = true;
							c++;
							if (allimages) cc = cmds.length;
						} else if (0 == cmds[cc].compareToIgnoreCase("/all")) {
							c++;
						} else cc = cmds.length;
						// else if (cmds[cc].startsWith("/")) cc = cmds.length;
					}
				} else if (!cmds[c].startsWith("/")) {
					boolean cfound = false;  // -- true if cmd[c] has been recognized as window or image title
					if (!allimages) {
						for (int i=0; i < imgLabels.length; i++)
							if (0 == imgLabels[i].compareToIgnoreCase(cmds[c])) {
								cfound = true;
								imgToClose[i] = true;
								break;
							}
						if (!cfound && 53 < cmds[c].length()) {
							String iCmd = cmds[c];
							while (!cfound && iCmd.endsWith("....")) {
								iCmd = iCmd.substring(0, iCmd.length()-1);
								for (int i=0; i < imgLabels.length; i++)
									if (0 == imgLabels[i].compareToIgnoreCase(iCmd)) {
										cfound = true;
										imgToClose[i] = true;
										break;
									}
							}
						}
						if (!cfound && 50 < cmds[c].length())
							for (int i=0; i<wList.length; i++) {
								String imgTitle = WindowManager.getImage(wList[i]).getTitle();
								if (50 < imgTitle.length()) {
									imgTitle = imgTitle.replaceAll(" ","_");
									if (0 == imgTitle.compareToIgnoreCase(cmds[c])) {
										cfound = true;
										imgToClose[i] = true;
										break;
									}
								}
							}
						if (!cfound && cmds[c].endsWith("_")) {
							String iCmd = cmds[c];
							do {
								iCmd = iCmd.substring(0, iCmd.length()-1);
								for (int i=0; i < imgLabels.length; i++)
									if (0 == imgLabels[i].compareToIgnoreCase(iCmd)) {
										cfound = true;
										imgToClose[i] = true;
										break;
									}
							} while (!cfound && iCmd.endsWith("_"));
						}
					} // -- if (!allimages)
					if (!cfound && !allframes) {
						for (int i=0; i < frameLabels.length; i++)
							if (0 == frameLabels[i].compareToIgnoreCase(cmds[c])) {
								cfound = true;
								winToClose[i] = true;
								break;
							}
						if (!cfound && 53 < cmds[c].length()) {
							String iCmd = cmds[c];
							while (!cfound && iCmd.endsWith("....")) {
								iCmd = iCmd.substring(0, iCmd.length()-1);
								for (int i=0; i < imgLabels.length; i++)
									if (0 == frameLabels[i].compareToIgnoreCase(iCmd)) {
										cfound = true;
										winToClose[i] = true;
										break;
									}
							}
						}
						if (!cfound && 50 < cmds[c].length()) {
							for (int i=0; i<frames.length; i++) {
								String winTitle = frames[i].getTitle();
								if (50 < winTitle.length()) {
									winTitle = winTitle.replaceAll(" ","_");
									if (0 == winTitle.compareToIgnoreCase(cmds[c])) {
										cfound = true;
										winToClose[i] = true;
										break;
									}
								}
							}
						}
						while (!cfound && cmds[c].endsWith("_")) {
							cmds[c] = cmds[c].substring(0, cmds[c].length()-1);
							for (int i=0; i < frameLabels.length; i++)
								if (0 == frameLabels[i].compareToIgnoreCase(cmds[c])) {
									cfound = true;
									winToClose[i] = true;
									break;
								}
						}
					} // -- if (!cfound && !allframes)
				} // -- else if (!cmds[c].startsWith("/"))
			} // -- while (++c < cmds.length)  if (0 < cmds[c].length())
			
			//  -----------------------------------------------------------------------------
			
		} else if (shiftKeyIsDown || altKeyIsDown) {
			// -- short user dialog modus
			//  		entered by pressing SHIFT or ALT key at plugin start
			//  		select disposeing for all images 
			//		and/or all non-image windows at once
			
			if (altKeyIsDown && reallySilentlyAtAltKey) {
				if (0 == wList.length) {
					// IJ.noImage();
					IJ.showStatus("No Image to dispose!");
					return;
				}
				allimages = true;
				// frames= new Frame[0];  // -- already set above
				// shouldBeKept = new boolean [0];
				// frameLabels = new String[0];
			} else {
				gd = new GenericDialog("Dispose ALL Windows");
				
				String plS = 1<wList.length ||  0<frames.length?"s":"";
				if (0<frames.length && 0<wList.length) {
					gd.addCheckbox("IMAGE windows without saving ("+wList.length+").", true);
					gd.addCheckbox("non-IMAGE windows without saving ("+frames.length+").", true);
				} else if (0<wList.length) {
					gd.addCheckbox("Image windows without saving ("+wList.length+").", true);
				} else if (0<frames.length) {
					gd.addCheckbox("Non-image windows without saving ("+frames.length+").", true);
				} else {
					IJ.showStatus("Nothing to dispose!");
					// IJ.noImage();
					return;
				}
				gd.showDialog();
				if (gd.wasCanceled()) return;
				if (Recorder.record)  Recorder.recordOption("/all");
				if (0<wList.length) {
					if (gd.getNextBoolean()) allimages = true;
					else  wList = new int[0];
				}
				if (0<frames.length) {
					if (gd.getNextBoolean()) allframes = true;
					else  frames= new Frame [0];
				}
			}
		} else {
			// -- long user dialog modus
			//  		select images and non-image windows to dispose
			
			int nbChkbxs = 5;
			// -- get maximal number of checkboxes shown at once in dialog,
			// 		depending on space available on screen
			if (nbChkbxs <  wList.length) nbChkbxs = getMaxNbChkbxs();
			else if (nbChkbxs < frames.length) nbChkbxs = getMaxNbChkbxs();
			
			//  -----------------------------------------------------------------------------
			
			if (0<wList.length) {
				//  --  select and close IMAGE windows  --
				int nbPages =  1 + (wList.length-1) / nbChkbxs;  // number of dialog pages required
					//		to confirm for all open image windows.  Each of these dialog 
					//		pages won't have more than (nbChkbxs) Checkboxes.
				int index = 0;
				for (int p=0; p<nbPages; p++) {
					String plS = 1<wList.length?"s":"";
					String eeS = 1<wList.length?" each":"";
					String[] items = {"Close selected image window"+plS+" without further confirmation.",
							"Close selected image window"+plS+" but only if changes are saved.",
							"Close selected image"+plS+" but confirm"+eeS+"  for saving changes before.",
							"Don't close any image window (skip to next page)"};
					if (!anyChanges) {
						items = new String[2];
						items[0] = "Close selected image window"+plS+" without saving";
						items[1] = "Don't close any image window (skip to next page)";
					}
					if (p==nbPages-1) {
						if (0<frames.length) items[items.length-1] = "Don't close any image (skip to non-image window page)";
						else if (1==nbPages) items[items.length-1] = "Don't close any image window (do nothing)";
					}
					
					// -- temporaily disable macro recording
					boolean isRcording = Recorder.record;
					if (isRcording) Recorder.record = false;
					gd = new GenericDialog("Select Image Window"+plS+" to Close");
					gd.addChoice("", items, items[index]);
					int ccnt = 0;  // -- number of added checkboxes
					for (int i=0; i<nbChkbxs; i++) {
						if (i+p*nbChkbxs < imgLabels.length) {
							String label = imgLabels[i+p*nbChkbxs];
							if (imgChanges[i]) label += " *";
							// -- don't select "Memory" window because the user presumeably wants to keep it,
							//		and don't select putatively erroreous list entries...
							gd.addCheckbox(label, null != imp[i+p*nbChkbxs] && 0 != label.compareTo("Memory") && 0 != label.compareTo("_"));
							ccnt++;
						}
					}
					if (1 < ccnt) {
						Panel ButtonPanel = new Panel();
						ButtonPanel.setLayout(new GridLayout(0,2, 15, 0));
						selectAllButton = new Button("select all");
						selectAllButton.addActionListener(gdal);
						deselectAllButton = new Button("deselect all");
						deselectAllButton.addActionListener(gdal);
						ButtonPanel.add(selectAllButton);
						ButtonPanel.add(deselectAllButton);
						gd.addPanel(new Panel());  // -- spacer
						gd.addPanel(ButtonPanel);
					}
					
					Panel MsgPanel = new Panel();
					MsgPanel.setLayout(new GridLayout(0,1));
					if (anyChanges) MsgPanel.add(new Label("Changed image"+plS+" are marked with * ."));
					MsgPanel.add(new Label("Run with <Shift> or <Alt> key pressed to skip this screen."));
					gd.addPanel(new Panel());  // -- spacer
					gd.addPanel(MsgPanel);
					
					// -- add a "Skip" button if it makes sense
					boolean yesNoCancelEnabled = false;
					if (1 < nbPages || 0 < frames.length) yesNoCancelEnabled = enableSkipButton();
					
					// -- show  dialog
					gd.showDialog();
					// -- re-enable recording (won't affect the GenericDialog that is being shown)
					if (isRcording)  Recorder.record = true;
					
					// -- read out user input
					boolean skipToNext = false;
					if(gd.wasCanceled()) {
						if (yesNoCancelEnabled || !dontQuitAtCancel) return;
						skipToNext = true;
					} else if (yesNoCancelEnabled) {
						try {
							skipToNext = !gd.wasOKed(); // -- true if "Skip" button had been pressed
						}
						catch (NoSuchMethodError e) {}
					}
					if (!skipToNext) {
						index = gd.getNextChoiceIndex();
						// -- skip if the user chose to do so
						skipToNext = index == items.length-1;
					}
					if (!skipToNext) {
						for (int i=0; i<nbChkbxs; i++) if (i+p*nbChkbxs < wList.length) {
							if (gd.getNextBoolean()) {
								if (0 < index) {
									if (!imgChanges[i+p*nbChkbxs]) {
										imgToClose[i+p*nbChkbxs] = null != imp[i+p*nbChkbxs];
									} else if (2 == index && null != imp[i+p*nbChkbxs]) {
										SaveChangesDialog d = new SaveChangesDialog(IJ.getInstance(), imgLabels[i+p*nbChkbxs]);
										if (d.savePressed()) imgToClose[i+p*nbChkbxs] = new FileSaver(imp[i+p*nbChkbxs]).save();
										else imgToClose[i+p*nbChkbxs] = !d.cancelPressed();
									} // else imgToClose[i+p*nbChkbxs]=false;
								} else imgToClose[i+p*nbChkbxs] = null != imp[i+p*nbChkbxs];
							}
						}
					}
					
					// -- clean up
					if (null != selectAllButton) {
						selectAllButton.removeActionListener(gdal);
						selectAllButton = null;
					}
					if (null != deselectAllButton) {
						deselectAllButton.removeActionListener(gdal);
						deselectAllButton = null;
					}
				}  // for (int p=0; p<nbPages; p++)
			} // if (0<wList.length)
			
			//  -----------------------------------------------------------------------------
			
			//  --  select NON-IMAGE windows
			if (0<frames.length) {
				int nbPages =  1 + (frames.length-1) / nbChkbxs;
				try {
					// optionalCloseAvailabe  = !IJ.versionLessThan("1.43s"); 
					optionalCloseAvailabe = ImageJ.VERSION.compareTo("1.43s") >= 0;
					// --- required for  TextWindow.close(boolean) method 
				}
				catch (NoSuchMethodError e) {}
				
				for (int p=0; p<nbPages; p++) {
					// -- temporaily disable macro recording
					boolean isRcording = Recorder.record;
					if (isRcording) Recorder.record = false;
					
					// -- generate user dialog for disposing non-image windows
					gd = new GenericDialog("Select non-Image Windows to Close");
					Choice methodChooser = new Choice();
					if (optionalCloseAvailabe) methodChooser.add("Close selected non-image windows, discard all text changes");
					methodChooser.add("Close selected non-image windows, confirm closing for changed/unsaved text");
					if (p+1< nbPages) methodChooser.add("Don't close any of these windows (skip to next page)");
					else methodChooser.add("Don't close any of these windows (skip)");
					Panel methodPanel = new Panel();
					methodPanel.add(methodChooser);
					gd.addPanel(methodPanel);
					int ccnt = 0;  // -- number of checkboxes added to dialog
					for (int i=0; i<nbChkbxs; i++) {
						if (i+p*nbChkbxs < frameLabels.length) {
							gd.addCheckbox(frameLabels[i+p*nbChkbxs], !shouldBeKept[i+p*nbChkbxs]);
							if (shouldBeKept[i+p*nbChkbxs] || !cbHeightMeasured) {
								java.util.Vector cb = gd.getCheckboxes();
								Checkbox cbx = (Checkbox) cb.get(ccnt);
								if (null != cbx) {
									// -- measure height of checkbox and save its value right away for next run
									if (!cbHeightMeasured) {
										checkboxHeight = (int) cbx.getSize().getHeight()+1;
										Prefs.set("checkboxHeight.int", (double) checkboxHeight);
										cbHeightMeasured = true;
									}
									cbx.setEnabled(!shouldBeKept[i+p*nbChkbxs]);
									if (gd.wasCanceled()) IJ.wait(200);  //  protecting against crash
								}
							}
							ccnt++;
						}
					}
					
					// -- add "Select all" and "Deselect all" Buttons
					if (1 < ccnt) {
						Panel ButtonPanel = new Panel();
						ButtonPanel.setLayout(new GridLayout(0,2, 15, 0));
						selectAllButton = new Button("select all");
						selectAllButton.addActionListener(gdal);
						deselectAllButton = new Button("deselect all");
						deselectAllButton.addActionListener(gdal);
						ButtonPanel.add(selectAllButton);
						ButtonPanel.add(deselectAllButton);
						gd.addPanel(new Panel());  // -- spacer
						gd.addPanel(ButtonPanel);
					}
					
					// -- add comments
					Panel MsgPanel = new Panel();
					MsgPanel.add(new Label("Run with <Shift> or <Alt> key pressed to skip this screen."));
					gd.addPanel(new Panel());  // -- spacer
					gd.addPanel(MsgPanel);
					
					// -- try to add a "Skip" button
					boolean yesNoCancelEnabled = false;
					if (1 < nbPages || 0 < wList.length) yesNoCancelEnabled = enableSkipButton();
					
					// -- show  the dialog
					gd.showDialog();
					// -- re-enable recording (won't affect the GenericDialog that is being shown)
					if (isRcording)  Recorder.record = true;
					
					// -- read out user input
					boolean skipToNext = false;
					if(gd.wasCanceled()) {
						if (yesNoCancelEnabled || !dontQuitAtCancel) return;
						skipToNext = true;
					} else if (yesNoCancelEnabled) {
						try {
							skipToNext = !gd.wasOKed(); // -- "Skip" button had been pressed
						}
						catch (NoSuchMethodError e) {}
					}
					if (!skipToNext) {
						int index = methodChooser.getSelectedIndex();
						if ((optionalCloseAvailabe && 2 > index) || 1 > index)
							for (int i=0; i<nbChkbxs; i++) 
								if (i+p*nbChkbxs < frames.length) {
									winToClose[i+p*nbChkbxs] = gd.getNextBoolean();
									confirmClosing[i+p*nbChkbxs] = optionalCloseAvailabe && 1 > index;
								}
					}
					
					// -- clean up
					if (null != selectAllButton) {
						selectAllButton.removeActionListener(gdal);
						selectAllButton = null;
					}
					if (null != deselectAllButton) {
						deselectAllButton.removeActionListener(gdal);
						deselectAllButton = null;
					}
				}  // for (int p=0; p<nbPages; p++)
			}
			
			//  -----------------------------------------------------------------------------
			
		} // -- if (macroRunningSilently)...; else if (shiftKeyIsDown || altKeyIsDown)...; else...
		
		if (allimages) for (int i=0; i < wList.length; i++) imgToClose[i] = true;
		if (allframes) for (int i=0; i < frames.length; i++) winToClose[i] = true;
		
		//  -----------------------------------------------------------------------------
		
		// --  close selected image windows
		int nClosedImages = 0;  // -- sucess statistics
		String resultsStr = "Disposed: "; // -- message to display in status bar when done
		if (0<wList.length) {
			IJ.showStatus("Disposing    image windows - please wait...");
			for (int i=0; i<wList.length; i++) if (imgToClose[i]){
				IJ.showStatus("Disposing "+(nClosedImages+1)+" image windows - please wait...");
				try {
					imp[i].unlock();
					imp[i].changes = false;
					ImageWindow win = imp[i].getWindow();
					if (null!=win) win.close();
					imp[i].trimProcessor();
					imp[i].flush();
					imp[i] = null;
					nClosedImages++;
				} catch (NullPointerException ex) {imgToClose[i] = false;}
				if (Recorder.record && !allimages && imgToClose[i]) {
					String key = imgLabels[i];
					key = Macro.trimKey(key); // -- removing "*" marks etc.
					Recorder.recordOption(key);
				}
			}
			resultsStr += nClosedImages + " of  " + wList.length + " image"; 
			if (0<frames.length) resultsStr += ", ";
			else resultsStr += " windows.";
		}
		
		//  -----------------------------------------------------------------------------
		
		// --  close non-selected image windows
		int nClosedFrames = 0;  // -- sucess statistics
		if (0<frames.length) {
			IJ.showStatus("Disposing    non-image windows...");
			for (int i=0; i<frames.length; i++) if (winToClose[i] && !shouldBeKept[i]) {
				IJ.showStatus("Disposing "+(nClosedFrames+1)+" non-image windows...");
				if (null != frames[i]) {
					if (frames[i] instanceof TextWindow)  {
						TextWindow tw = (TextWindow) frames[i];
						if (optionalCloseAvailabe) {
							try {
								tw.close(confirmClosing[i]);
							}
							catch (NoSuchMethodError e) {
								// -- actually this should't be the case, however:
								optionalCloseAvailabe  = false;
								tw.close();
							}
						} else tw.close();
					} else {
						WindowManager.removeWindow(frames[i]);
						frames [i].dispose();
					}
					nClosedFrames++;
					if (Recorder.record && !allframes) {
						String key = frameLabels[i];
						key = Macro.trimKey(key);
						Recorder.recordOption(key);
					}
				}
			}
			resultsStr += nClosedFrames + " of  " + frames.length + " non-image windows."; 
		} else if (0 == wList.length) resultsStr += "nothing!";
		
		//  -----------------------------------------------------------------------------
		
		// -- run the garbage collector
		IJ.showStatus("Running the garbage collector...");
		if (0 < nClosedImages && 0 == nClosedFrames) IJ.wait(200);   // ###
		System.gc();
		// -- update Window Manager
		try {
			WindowManager.repaintImageWindows();
		}
 		catch (NoSuchMethodError ex) {}
		// -- display success statistics
		IJ.showStatus(resultsStr);
		// IJ.wait(3000);
		// IJ.showStatus("");
	} // run()
	
	private boolean enableSkipButton() {
		boolean success = false;
		final String OKButtonLabel = "  OK  ";
		final String NoButtonLabel = " Skip ";
		try {
			gd.enableYesNoCancel(OKButtonLabel, NoButtonLabel);
			success = true;
		}
		catch (NoSuchMethodError e) {}
		if (success) return true;
		
		try {
			gd.enableYesNoCancel();
			success = true;
			gd.setOKLabel(OKButtonLabel);
		}
		catch (NoSuchMethodError e) {}
		return success;
	}
	
	private final int getMaxNbChkbxs() {
		// -- get checkbox height
		if (!cbHeightMeasured) checkboxHeight = (int)Prefs.get("checkboxHeight.int", 0);
		if (0 != checkboxHeight) cbHeightMeasured = true;
		else checkboxHeight = 2 * IJ.getInstance().getFont().getSize();  
				// -- otherwise take default value, two times font size, typically 24 pixels
		if (22 > checkboxHeight) checkboxHeight = 22;  // -- minimal value that could be right
		
		// -- get screen height
		GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
		Rectangle bounds = ge.getMaximumWindowBounds();
		int screenHeight = bounds.height;
		
		// -- subtract from screen height some space for other dialog elements to display
		int spaceOutside = 10;  // -- free space outside the dialog: 150 px  (change this setting if you want)
		Dimension ijSize = ij!=null?ij.getSize():new Dimension(0,0);
		if (screenHeight > 600)  spaceOutside += ijSize.height;
		int dialogSizeWithoutCheckBoxes = 120 + 5 * checkboxHeight;  // -- considering 5 lines of text in labels, choice and buttons
		int freeCanvasHeight = screenHeight - dialogSizeWithoutCheckBoxes - spaceOutside;
		
		int NbChkbxs = (int) Math.ceil((double) freeCanvasHeight / checkboxHeight);
		if (4 > NbChkbxs) NbChkbxs = 4;
		return NbChkbxs;
	}
	
}